from pathlib import Path

import pytest  # type: ignore
from ape import Project, networks
from ape_http.providers import EthereumNetworkConfig

from ape_hardhat import HardhatProvider


def get_project():
    return Project(Path(__file__).parent)


def get_network_config():
    p = get_project()
    config_classes = [
        klass for (name, klass) in p.config.plugin_manager.config_class if name == "hardhat"
    ]
    assert len(config_classes) == 1
    config = config_classes[0]()

    # need to instantiate a new instance of this otherwise it's shared across HH instances
    config.ethereum = EthereumNetworkConfig()

    # bump up the timeouts to decrease chance of tests flaking due to race conditions
    config.network_retries = [0.1, 0.2, 0.3, 0.5, 0.5, 1, 1, 1, 1, 1, 1, 5]
    config.process_attempts = 10

    return config


@pytest.fixture
def project():
    return get_project()


@pytest.fixture
def network_api():
    return networks.ecosystems["ethereum"]["development"]


@pytest.fixture
def network_config(project):
    return get_network_config()


@pytest.fixture
def hardhat_provider(network_api, network_config):
    hh = HardhatProvider("hardhat", network_api, network_config, {}, Path("."), "")
    hh.connect()
    return hh



import time
from pathlib import Path

import pytest  # type: ignore
from conftest import get_network_config
from hexbytes import HexBytes

from ape_hardhat import HardhatProvider, HardhatProviderError

TEST_WALLET_ADDRESS = "0xD9b7fdb3FC0A0Aa3A507dCf0976bc23D49a9C7A3"
TEST_CUSTOM_PORT = 8555  # vs. Hardhat's default of 8545
TEST_CUSTOM_URI = f"http://localhost:{TEST_CUSTOM_PORT}"


def test_instantiation(network_api, network_config):
    h = HardhatProvider("hardhat", network_api, network_config, {}, Path("."), "")
    assert h.name == "hardhat"


def test_connect(network_api, network_config):
    h = HardhatProvider("hardhat", network_api, network_config, {}, Path("."), "")
    h.connect()
    assert h.chain_id == 31337


def test_disconnect(hardhat_provider):
    p = hardhat_provider.process
    assert p
    hardhat_provider.disconnect()
    for i in range(15):
        if hardhat_provider.process is None and p.poll() is not None:
            return True  # this means the process exited
        time.sleep(0.1)
    raise RuntimeError("hardhat process didn't exit in time")


def test_gas_price(hardhat_provider):
    gas_price = hardhat_provider.gas_price
    assert gas_price > 1


def test_uri(hardhat_provider):
    assert f"http://localhost:{hardhat_provider.port}" in hardhat_provider.uri


@pytest.mark.parametrize(
    "method,args,expected",
    [
        (HardhatProvider.get_nonce, [TEST_WALLET_ADDRESS], 0),
        (HardhatProvider.get_balance, [TEST_WALLET_ADDRESS], 0),
        (HardhatProvider.get_code, [TEST_WALLET_ADDRESS], HexBytes("")),
    ],
)
def test_rpc_methods(hardhat_provider, method, args, expected):
    assert method(hardhat_provider, *args) == expected


def test_custom_port(network_api, network_config):
    network_config.port = TEST_CUSTOM_PORT
    network_config.ethereum.development["uri"] = TEST_CUSTOM_URI
    h = HardhatProvider("hardhat", network_api, network_config, {}, Path("."), "")
    h.connect()
    assert TEST_CUSTOM_URI in h.uri
    assert h.config.port == TEST_CUSTOM_PORT


def test_two_hardhat_instances(network_api):
    """
    Validate the somewhat tricky internal logic of running two Hardhat subprocesses
    under a single parent process.
    """
    # configure the first provider with a custom port
    nc1 = get_network_config()
    nc1.port = TEST_CUSTOM_PORT
    nc1.ethereum.development["uri"] = TEST_CUSTOM_URI

    # configure the second provider with the default port
    nc2 = get_network_config()

    # instantiate the providers (which will start the subprocesses) and validate the ports
    h1 = HardhatProvider("hardhat", network_api, nc1, {}, Path("."), "")
    h2 = HardhatProvider("hardhat", network_api, nc2, {}, Path("."), "")
    h1.connect()
    h2.connect()

    # validate the two instances, subprocesses, and blockchains are independent of each other
    assert TEST_CUSTOM_URI in h1.uri
    assert h1.config.port == h1.port == TEST_CUSTOM_PORT
    # the web3 clients must be different in the HH provider instances (compared to the
    # behavior of the EthereumProvider base class, where it's a shared classvar)
    assert h1._web3 != h2._web3
    assert h2.port > h1.port  # h2 will have a higher port number in the ephemeral port range
    h1.mine()
    h2.mine()
    hash1 = h1._web3.eth.get_block("latest").hash
    hash2 = h2._web3.eth.get_block("latest").hash
    assert hash1 != hash2


def test_set_block_gas_limit(hardhat_provider):
    gas_limit = hardhat_provider._web3.eth.get_block("latest").gasLimit
    assert hardhat_provider.set_block_gas_limit(gas_limit) is True


def test_sleep(hardhat_provider):
    seconds = 5
    t1 = hardhat_provider.sleep(seconds)
    t2 = hardhat_provider.sleep(seconds)
    assert t2 - t1 == seconds


def test_mine(hardhat_provider):
    block1 = hardhat_provider._web3.eth.get_block("latest")
    assert hardhat_provider.mine() == "0x0"
    block2 = hardhat_provider._web3.eth.get_block("latest")
    assert hardhat_provider.mine() == "0x0"
    assert block1.hash != block2.hash and block2.number > block1.number


def test_revert_failure(hardhat_provider):
    assert hardhat_provider.revert(0xFFFF) is False


def test_snapshot_and_revert(hardhat_provider):
    snap = hardhat_provider.snapshot()
    assert snap == 1
    block1 = hardhat_provider._web3.eth.get_block("latest")
    hardhat_provider.mine()
    block2 = hardhat_provider._web3.eth.get_block("latest")
    assert block1.hash != block2.hash and block2.number > block1.number
    hardhat_provider.revert(snap)
    block3 = hardhat_provider._web3.eth.get_block("latest")
    assert block1.hash == block3.hash and block1.number == block3.number


def test_unlock_account(hardhat_provider):
    assert hardhat_provider.unlock_account(TEST_WALLET_ADDRESS) is True


def test_double_connect(hardhat_provider):
    # connect has already been called once as part of the fixture, so connecting again should fail
    with pytest.raises(HardhatProviderError):
        hardhat_provider.connect()



"""
Implementation for HardhatProvider.
"""

import atexit
import ctypes
import platform
import random
import shutil
import signal
import sys
import time
from subprocess import PIPE, Popen, call
from typing import Any, List, Optional

from ape.exceptions import ProviderError
from ape.logging import logger
from ape_http.providers import DEFAULT_SETTINGS, EthereumProvider, NetworkConfig

EPHEMERAL_PORTS_START = 49152
EPHEMERAL_PORTS_END = 60999
HARDHAT_CHAIN_ID = 31337
HARDHAT_CONFIG = """
// See https://hardhat.org/config/ for config options.
module.exports = {
  networks: {
    hardhat: {
      hardfork: "london",
      // base fee of 0 allows use of 0 gas price when testing
      initialBaseFeePerGas: 0,
    },
  },
};
"""
HARDHAT_START_NETWORK_RETRIES = [0.1, 0.2, 0.3, 0.5, 1.0]  # seconds between network retries
HARDHAT_START_PROCESS_ATTEMPTS = 3  # number of attempts to start subprocess before giving up
PROCESS_WAIT_TIMEOUT = 15  # seconds to wait for process to terminate


def _signal_handler(signum, frame):
    """Runs on SIGTERM and SIGINT to force ``atexit`` handlers to run."""
    atexit._run_exitfuncs()
    sys.exit(143 if signum == signal.SIGTERM else 130)


def _linux_set_death_signal():
    """
    Automatically sends SIGTERM to child subprocesses when parent process
    dies (only usable on Linux).
    """
    # from: https://stackoverflow.com/a/43152455/75956
    # the first argument, 1, is the flag for PR_SET_PDEATHSIG
    # the second argument is what signal to send to child subprocesses
    libc = ctypes.CDLL("libc.so.6")
    return libc.prctl(1, signal.SIGTERM)


def _get_preexec_fn():
    if platform.uname().system == "Linux":
        return _linux_set_death_signal
    else:
        return None


def _windows_taskkill(pid: int) -> None:
    """
    Kills the given process and all child processes using taskkill.exe. Used
    for subprocesses started up on Windows which run in a cmd.exe wrapper that
    doesn't propagate signals by default (leaving orphaned processes).
    """
    taskkill_bin = shutil.which("taskkill")
    if not taskkill_bin:
        raise HardhatSubprocessError("Could not find taskkill.exe executable.")
    proc = Popen(
        [
            taskkill_bin,
            "/F",  # forcefully terminate
            "/T",  # terminate child processes
            "/PID",
            str(pid),
        ],
        stderr=PIPE,
        stdout=PIPE,
        stdin=PIPE,
    )
    proc.wait(timeout=PROCESS_WAIT_TIMEOUT)


def _kill_process(proc):
    """Helper function for killing a process and its child subprocesses."""
    if platform.uname().system == "Windows":
        _windows_taskkill(proc.pid)
    proc.kill()
    proc.wait(timeout=PROCESS_WAIT_TIMEOUT)


class HardhatProviderError(ProviderError):
    """An error related to the Hardhat network provider plugin."""


class HardhatSubprocessError(HardhatProviderError):
    """An error related to launching subprocesses to run Hardhat."""


class HardhatNetworkConfig(NetworkConfig):
    # --port <INT, default from Hardhat is 8545, but our default is to assign a random port number>
    port: Optional[int] = None

    # retry strategy configs, try increasing these if you're getting HardhatSubprocessError
    network_retries: List[float] = HARDHAT_START_NETWORK_RETRIES
    process_attempts: int = HARDHAT_START_PROCESS_ATTEMPTS


class HardhatProvider(EthereumProvider):
    def __post_init__(self):
        self._hardhat_web3 = (
            None  # we need to maintain a separate per-instance web3 client for Hardhat
        )
        self.port = None
        self.process = None
        self.npx_bin = shutil.which("npx")

        hardhat_config_file = self.network.config_manager.PROJECT_FOLDER / "hardhat.config.js"
        if not hardhat_config_file.is_file():
            hardhat_config_file.write_text(HARDHAT_CONFIG)

        if not self.npx_bin:
            raise HardhatSubprocessError(
                "Could not locate NPM executable. See ape-hardhat README for install steps."
            )
        if call([self.npx_bin, "--version"], stderr=PIPE, stdout=PIPE, stdin=PIPE) != 0:
            raise HardhatSubprocessError(
                "NPM executable returned error code. See ape-hardhat README for install steps."
            )
        if call([self.npx_bin, "hardhat", "--version"], stderr=PIPE, stdout=PIPE, stdin=PIPE) != 0:
            raise HardhatSubprocessError(
                "Missing hardhat NPM package. See ape-hardhat README for install steps."
            )

        # register atexit handler to make sure disconnect is called for normal object lifecycle
        atexit.register(self.disconnect)

        # register signal handlers to make sure atexit handlers are called when the parent python
        # process is killed
        signal.signal(signal.SIGINT, _signal_handler)
        signal.signal(signal.SIGTERM, _signal_handler)

    def _start_process(self):
        """Start the hardhat process and wait for it to respond over the network."""
        cmd = [self.npx_bin, "hardhat", "node"]

        # pick a random port if one isn't configured
        self.port = self.config.port
        if not self.port:
            self.port = random.randint(EPHEMERAL_PORTS_START, EPHEMERAL_PORTS_END)
        cmd.extend(["--port", str(self.port)])

        # TODO: Add configs to send stdout to logger / redirect to a file in plugin data dir?
        process = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, preexec_fn=_get_preexec_fn())
        connected = False
        for retry_time in self.config.network_retries:
            time.sleep(retry_time)
            super().connect()
            connected = self._verify_connection()
            if connected:
                break
        if not connected:
            if process.poll() is None:
                raise HardhatSubprocessError(
                    "Hardhat process is running, but could not connect to RPC server. "
                    "Run `npx hardhat node` or adjust retry strategy configs to troubleshoot."
                )
            raise HardhatSubprocessError(
                "Hardhat command exited prematurely. Run `npx hardhat node` to troubleshoot or "
                "adjust retry strategy configs to troubleshoot."
            )
        elif process.poll() is not None:
            raise HardhatSubprocessError(
                "Hardhat process exited prematurely, but connection succeeded. "
                "Is something else listening on the port?"
            )
        self.process = process

    def _verify_connection(self) -> bool:
        """Make a network call for chain_id and verify the result."""
        try:
            chain_id = self.chain_id
            if chain_id != HARDHAT_CHAIN_ID:
                raise HardhatProviderError(f"Unexpected chain ID: {chain_id}")
            return True
        except Exception as exc:
            logger.debug("Hardhat connection failed: %r", exc)
        return False

    def connect(self):
        """Start the hardhat process and verify it's up and accepting connections."""

        if self.process:
            raise HardhatProviderError(
                "Cannot connect twice. Call disconnect before connecting again."
            )

        if self.config.port:
            # if a port is configured, only make one start up attempt
            self._start_process()
        else:
            for _ in range(self.config.process_attempts):
                try:
                    self._start_process()
                    break
                except HardhatSubprocessError as exc:
                    logger.info("Retrying hardhat subprocess startup: %r", exc)

        # subprocess should be running and receiving network requests at this point
        if not (self.process and self.process.poll() is None and self.port):
            raise HardhatSubprocessError(
                "Could not start hardhat subprocess on a random port. "
                "See logs or run `npx hardhat node` or adjust retry strategy configs to "
                "troubleshoot."
            )

    @property
    def uri(self) -> str:
        uri = getattr(self.config, self.network.ecosystem.name)[self.network.name]["uri"]
        if uri != DEFAULT_SETTINGS["uri"]:
            # the user configured their own URI in the project configs, let's use that
            return uri
        elif self.port:
            # the user did not override the default URI, and we have a port
            # number, so let's build the URI using that port number
            return f"http://localhost:{self.port}"
        else:
            raise HardhatProviderError("Can't build URI before `connect` is called.")

    @property  # type: ignore
    def _web3(self):
        """
        This property overrides the ``EthereumProvider._web3`` class variable to return our
        instance variable.
        """
        return self._hardhat_web3

    @_web3.setter
    def _web3(self, value):
        """
        Redirect the base class's assignments of self._web3 class variable to our instance variable.
        """
        self._hardhat_web3 = value

    def disconnect(self):
        super().disconnect()
        if self.process:
            _kill_process(self.process)
        self.process = None
        self.port = None

    def _make_request(self, rpc: str, args: list) -> Any:
        return self._web3.manager.request_blocking(rpc, args)  # type: ignore

    def set_block_gas_limit(self, gas_limit: int) -> bool:
        return self._make_request("evm_setBlockGasLimit", [hex(gas_limit)])

    def sleep(self, seconds: int) -> int:
        return int(self._make_request("evm_increaseTime", [seconds]))

    def mine(self, timestamp: Optional[int] = None) -> str:
        return self._make_request("evm_mine", [timestamp] if timestamp else [])

    def snapshot(self) -> int:
        return self._make_request("evm_snapshot", [])

    def revert(self, snapshot_id: int) -> bool:
        return self._make_request("evm_revert", [snapshot_id])

    def unlock_account(self, address: str) -> bool:
        return self._make_request("hardhat_impersonateAccount", [address])
